@@ -1,4 +1,4 @@
-# Copyright 2013, 2014 Reahl Software Services (Pty) Ltd. All rights reserved.
+# Copyright 2013-2016 Reahl Software Services (Pty) Ltd. All rights reserved.
 #
 #    This file is part of Reahl.
 #
@@ -16,8 +16,12 @@
 
 from __future__ import print_function, unicode_literals, absolute_import, division
 import six
-from threading import Event
-from threading import Thread
+
+import os
+import time
+import subprocess
+import atexit
+from threading import Event, Thread, Timer
 import select
 from wsgiref import simple_server
 import sys
@@ -28,16 +32,26 @@ from contextlib import contextmanager
 import logging
 import functools
 import pkg_resources
+import warnings
+import distutils
 
 from webob import Request
 from webob.exc import HTTPInternalServerError
 
+from six.moves.queue import Queue
+
+from watchdog.observers import Observer
+from watchdog.events import PatternMatchingEventHandler
+from watchdog.observers.polling import PollingObserver
+
 from reahl.component.exceptions import ProgrammerError
 from reahl.component.context import ExecutionContext
 from reahl.component.config import StoredConfiguration
 from reahl.component.py3compat import ascii_as_bytes_or_str
+from reahl.component.shelltools import Executable
 from reahl.web.fw import ReahlWSGIApplication
 
+
 class WrappedApp(object):
     """A class in which to wrap a WSGI app, allowing catching of exceptions in the wrapped app.
     """
@@ -57,17 +71,10 @@ class WrappedApp(object):
             to_return = b''
             for i in app(environ, start_response):
                 to_return += i 
-        except socket.error:
-            to_return = b''
-            for i in HTTPInternalServerError(charset=ascii_as_bytes_or_str('utf-8'))(environ, start_response):
-                to_return += i
         except:
             to_return = b''
             (_, self.exception, self.traceback) = sys.exc_info()
-            try:
-                traceback_html = six.text_type(traceback.format_exc(self.traceback))
-            except Exception as ex:
-                traceback_html = 'Exception happened while trying to render traceback: %s' % ex
+            traceback_html = six.text_type(traceback.format_exc())
             for i in HTTPInternalServerError(content_type=ascii_as_bytes_or_str('text/plain'), charset=ascii_as_bytes_or_str('utf-8'), unicode_body=traceback_html)(environ, start_response):
                 to_return += i
         yield to_return
@@ -104,82 +111,133 @@ class NoopApp(object):
     def clear_exception(self):
         pass
 
+from wsgiref.simple_server import ServerHandler
 
-class LoggingRequestHandler(simple_server.WSGIRequestHandler):
+class PatchedServerHandler(ServerHandler):
+    # We modify simple_server.ServerHandler to work around some
+    # problems experienced with it.
+    def finish_response(self):
+        # If the browser closes the connection while we still want to sen stuff back,
+        # we want to fail silently and give up. This often happens in tests where the
+        # browser may want to request embedded links (like stylesheets) too, yet the
+        # test has already clicked on the next link.
+        if six.PY3:
+            ssl_eof_error = ssl.SSLEOFError
+            broken_pipe_error = BrokenPipeError
+        else:
+            ssl_eof_error = ssl.SSLError
+            broken_pipe_error = socket.error
+            
+        try:
+            ServerHandler.finish_response(self)
+        except (ssl_eof_error,  broken_pipe_error):
+            # Silently ignore it if it looks like the client browser closed the connection.
+            pass
+
+
+class SingleWSGIRequestHandler(simple_server.WSGIRequestHandler):
+    # We modify simple_server.WSGIRequestHandler slightly to work around some
+    # problems experienced with it.
+    def handle(self):
+        # A modified copy of simple_server.WSGIRequestHandler.handle
+        #
+        # We first check if there really IS something to read on the socket
+        # because as soon as ReahlWSGIServer.connection_is_pending becomes
+        # true, A SingleWSGIRequestHandler is created already. However, the
+        # fact that the browser made a connection does not mean it has sent
+        # HTTP requests yet. So, we cannot really serve yet.
+        #
+        # We first check whether something has actually arrived here by
+        # temporarily setting a timeout on the socket to read the first byte.
+        # We suspect that the browser connects eagerly (pre-connect) before it
+        # has a request to send in order to gain a speed benefit. 
+        #
+        # We also use this override to use our PatchedServerHandler instead of
+        # ServerHandler to handle HTTP requests.
+        try:
+            self.request.settimeout(0.1)
+            self.raw_requestline = self.rfile.read(1)
+        except (socket.timeout, ssl.SSLError):
+            return
+        finally:
+            self.request.settimeout(None)
+            
+        self.raw_requestline = self.raw_requestline + self.rfile.readline(65536)
+        if len(self.raw_requestline) > 65536:
+            self.requestline = ''
+            self.request_version = ''
+            self.command = ''
+            self.send_error(414)
+            return
+
+        if not self.parse_request(): # An error code has been sent, just exit
+            return
+
+        handler = PatchedServerHandler(
+            self.rfile, self.wfile, self.get_stderr(), self.get_environ()
+        )
+        handler.request_handler = self      # backpointer for logging
+        handler.run(self.server.get_app())
+
+        
     def log_message(self, format, *args):
         message = format % args
         logging.getLogger(__name__).info(message)
 
-    def setup(self):
-        simple_server.WSGIRequestHandler.setup(self)
-        # The server will get here if requests_waiting() returned True.
-        # In some cases (ref: chrome prefetching/preconnecting) requests_waiting() will return True,
-        #   but there really is nothing to read on the socket, and our server will block here.
-        # All this timeout stuff justs safeguards against that eventuality - which is a hard thing
-        #   to debug if it happens.
-        if six.PY2:
-            self.rfile._sock.settimeout(3)  
-        else:
-            self.connection.settimeout(3)
 
-    def handle(self):
-        try:
-            simple_server.WSGIRequestHandler.handle(self)
-        except socket.timeout:
-            message = 'Server socket timed out waiting to receive the request. This may happen if the server mistakenly deduced that there were requests waiting for it when there were not. Such as when chrome prefetches things, etc.'
-            logging.getLogger(__name__).warn(message)
-
-    def finish_response(self):
-        simple_server.WSGIRequestHandler.finish_response()
-
-
-class SSLWSGIRequestHandler(LoggingRequestHandler):
-    def get_environ(self):
-        env = simple_server.WSGIRequestHandler.get_environ(self)
-        env['HTTPS']='on'
-        return env
 
 
 class ReahlWSGIServer(simple_server.WSGIServer):
     @classmethod
     def make_server(cls, host, port, reahl_wsgi_app):
         httpd = simple_server.make_server(host, port, reahl_wsgi_app, server_class=cls, 
-                                          handler_class=LoggingRequestHandler)
+                                          handler_class=SingleWSGIRequestHandler)
         return httpd
 
     def __init__(self, server_address, RequestHandlerClass):
         simple_server.WSGIServer.__init__(self, server_address, RequestHandlerClass)
         self.allow_reuse_address = True
         
-    def serve_async(self):
-        if self.requests_waiting(0.01):
-            self.handle_waiting_request()
+    def serve_async(self, in_separate_thread=False):
+        if self.connection_is_pending(0.01):
+            self.handle_waiting_request(in_separate_thread)
 
-    def handle_waiting_request(self):
+    def handle_waiting_request(self, in_separate_thread):
+        self.in_separate_thread = in_separate_thread
         self.handle_request()
         try:
             self.get_app().report_exception()
         finally:
             self.get_app().clear_exception()
     
-    def requests_waiting(self, timeout):
-        i, o, w = select.select([self.socket],[],[],timeout)
+    def connection_is_pending(self, timeout):
+        i, o, w = select.select([self],[],[],timeout)
         return i
 
+    def handle_error(self, request, client_address):
+        exc_type, exc, tb = sys.exc_info()
+        if isinstance(exc, socket.error) and exc.errno == 32:
+            # PY2: Silently ignore it if it looks like the client browser closed the socket.
+            pass
+        else:
+            simple_server.WSGIServer.handle_error(self, request, client_address)
+                
+
 
 class SSLCapableWSGIServer(ReahlWSGIServer):
     @classmethod
     def make_server(cls, host, port, certfile, reahl_wsgi_app):
         cls.certfile = certfile
         httpd = simple_server.make_server(host, port, reahl_wsgi_app, server_class=cls, 
-                                          handler_class=SSLWSGIRequestHandler)
+                                          handler_class=SingleWSGIRequestHandler)
         return httpd
 
-    def __init__(self, server_address, RequestHandlerClass):
-        ReahlWSGIServer.__init__(self, server_address, RequestHandlerClass)
-        self.socket = ssl.wrap_socket(self.socket, certfile=self.certfile)
-        
-    def get_request(self):
+    def setup_environ(self):
+        ReahlWSGIServer.setup_environ(self)
+        self.base_environ['HTTPS']='on'
+
+    def server_bind(self):
+        self.socket = ssl.wrap_socket(self.socket, server_side=True, certfile=self.certfile)
         if six.PY2:
             #This method is shamelessly copied from WerkZeug (and changed)
             class _SSLConnectionFix(object):
@@ -194,21 +252,17 @@ class SSLCapableWSGIServer(ReahlWSGIServer):
 
                 def __getattr__(self, attrib):
                     return getattr(self._con, attrib)
+                
+            old_accept = self.socket.accept
+            def patched_accept():
+                con, info = old_accept()
+                return _SSLConnectionFix(con), info
+            self.socket.accept = patched_accept
+        ReahlWSGIServer.server_bind(self)
 
-            con, info = self.socket.accept()
-            con = _SSLConnectionFix(con)
-        else:
-            con, info = self.socket.accept()
-        
-        return con, info
 
-    def finish_request(self, request, client_address):
-        try:
-            ReahlWSGIServer.finish_request(self, request, client_address)
-        except ssl.SSLError as ex:
-            pass
 
-class Handler(object):
+class WebDriverHandler(object):
     def __init__(self, command_executor):
         self.command_executor = command_executor
         self.original_execute = command_executor.execute
@@ -249,6 +303,124 @@ class Handler(object):
         self.command_executor.execute = wrapped_execute
 
 
+class Py2TimeoutExpired(Exception):
+    pass
+
+class SlaveProcess(object):
+    def __init__(self, command, args):
+        self.process = None
+        self.args = args
+        self.executable = Executable(command)
+
+    def terminate(self, timeout=5):
+        logging.getLogger(__name__).debug('Terminating process with PID[%s]' % self.process.pid)
+        self.process.terminate()
+        self.wait_to_die(timeout=timeout)
+
+    def wait_to_die(self, timeout):
+        TimeoutExpired = Py2TimeoutExpired if six.PY2 else subprocess.TimeoutExpired
+        try:
+            if six.PY2:
+                self.py2_process_wait_within_timeout(timeout)
+            else:
+                self.process.wait(timeout=timeout)
+        except TimeoutExpired:
+            self.process.kill()
+
+    def py2_process_wait_within_timeout(self, timeout):
+        thread = Thread(target=self.process.wait)
+        thread.start()
+        thread.join(timeout)
+        if thread.isAlive():
+            raise Py2TimeoutExpired()
+
+    def spawn_new_process(self):
+        return self.executable.Popen(self.args, env=os.environ.copy())
+
+    def is_running(self):
+        return self.process.poll() is None
+
+    def start(self):
+        self.process = self.spawn_new_process()
+        self.register_orphan_killer(self.create_orphan_killer(self.process))
+        logging.getLogger(__name__).debug('Starting process with PID[%s]' % self.process.pid)
+
+    def create_orphan_killer(self, process):
+        def kill_orphan_on_exit(possible_orphan_process):
+            logging.getLogger(__name__).debug('Cleanup: ensuring process with PID[%s] has terminated' % possible_orphan_process.pid)
+            try:
+                possible_orphan_process.kill()
+                logging.getLogger(__name__).debug('Had to kill process(orphan) with PID[%s]' % possible_orphan_process.pid)
+            except (OSError if six.PY2 else ProcessLookupError):
+                logging.getLogger(__name__).debug('Process with PID[%s] seems terminated already, no need to kill it' % possible_orphan_process.pid)
+        return functools.partial(kill_orphan_on_exit, process)
+
+    def register_orphan_killer(self, kill_function):
+        atexit.register(kill_function)
+
+    def restart(self):
+        self.terminate()
+        self.start()
+
+
+class ServerSupervisor(PatternMatchingEventHandler):
+    def __init__(self, slave_process_args, max_seconds_between_restarts, directories_to_monitor=['.']):
+        super(ServerSupervisor, self).__init__(ignore_patterns=['*.pyc','*.pyo', '*/__pycache__/*'], ignore_directories=True)
+        self.serving_process = SlaveProcess(sys.argv[0], slave_process_args)
+        self.max_seconds_between_restarts = max_seconds_between_restarts
+        self.directories_to_monitor = directories_to_monitor
+        self.directory_observers = []
+        self.files_changed = Event()
+        self.request_stop_running = Event()
+
+    #override for PatternMatchingEventHandler
+    def on_any_event(self, event):
+        self.files_changed.set()
+
+    def start_observing_directories(self):
+        for directory in self.directories_to_monitor:
+            directory_observer = Observer()
+            directory_observer.schedule(self, directory, recursive=True)
+            directory_observer.start()
+            self.directory_observers.append(directory_observer)
+
+    def stop_observing_directories(self):
+        for directory_observer in self.directory_observers:
+            directory_observer.stop()
+            directory_observer.join()
+
+    def pause(self):
+        time.sleep(self.max_seconds_between_restarts)
+
+    def stop(self):
+        self.request_stop_running.set()
+
+    def stop_serving(self):
+        if self.serving_process.is_running():
+            self.serving_process.terminate()
+        self.stop_observing_directories()
+
+    def run(self):
+        self.files_changed.clear()
+        self.start_observing_directories()
+        self.serving_process.start()
+
+        while not self.request_stop_running.is_set():
+            try:
+                self.pause()
+                if self.files_changed.is_set():
+                    logging.getLogger(__name__).debug('Changes to filesystem detected, scheduling a restart...')
+                    self.files_changed.clear()
+                    if self.serving_process.is_running():
+                        self.serving_process.restart()
+                    else:
+                        self.serving_process.start()
+            except KeyboardInterrupt:
+                self.stop()
+
+        self.stop_serving()
+
+
 class ReahlWebServer(object):
     """A web server for testing purposes. This web server runs both an HTTP and HTTPS server. It can 
        be configured to handle requests in the same thread as the test itself, but it can also be run in a
@@ -277,7 +449,8 @@ class ReahlWebServer(object):
         self.set_app(NoopApp())
 
     def __init__(self, config, port):
-        self.in_seperate_thread = None
+        super(ReahlWebServer, self).__init__()
+        self.in_separate_thread = None
         self.running = False
         self.handlers = {}
         self.httpd_thread = None
@@ -293,11 +466,19 @@ class ReahlWebServer(object):
                       
             raise AssertionError(message)
 
+
     def main_loop(self, context):
         with context:
             while self.running:
-                self.httpd.serve_async()
-                self.httpsd.serve_async()
+                try:
+                    self.httpd.serve_async(in_separate_thread=self.in_separate_thread)
+                    self.httpsd.serve_async(in_separate_thread=self.in_separate_thread)
+                except:  
+                    # When running as a standalone server, we keep the server running, but else break so tests break
+                    if self.in_separate_thread and self.running:
+                        print(traceback.format_exc(), file=sys.stderr)
+                    else:
+                        raise
 
     def start_thread(self):
         assert not self.running
@@ -314,16 +495,31 @@ class ReahlWebServer(object):
                 raise ProgrammerError('Timed out after 5 seconds waiting for httpd serving thread to end')
         self.httpd_thread = None
 
-    def start(self, in_seperate_thread=True, connect=False):
+    def wait_for_server_to_complete(self):
+        try:
+            self.httpd_thread.join()
+        finally:
+            self.stop()
+
+    def start(self, in_separate_thread=True,  connect=False, in_seperate_thread=None):
         """Starts the webserver and web application.
         
-           :keyword in_seperate_thread: If False, the server handles requests in the same thread as your tests.
+           :keyword in_separate_thread: If False, the server handles requests in the same thread as your tests.
+           :keyword in_seperate_thread: Deprecated: rather use in_separate_thread keyword argument
            :keyword connect: If True, also connects to the database.
         """
         self.reahl_wsgi_app.start(connect=connect)
+
         if in_seperate_thread:
+            warnings.warn('The in_seperate_thread keyword argument is deprecated, please use in_separate_thread instead.',
+               DeprecationWarning, stacklevel=2)
+            self.in_separate_thread = in_seperate_thread
+        else:
+            self.in_separate_thread = in_separate_thread
+
+        if self.in_separate_thread:
             self.start_thread()
-        self.in_seperate_thread = in_seperate_thread
+
 
     def stop(self):
         """Stops the webserver and web application from running."""
@@ -332,26 +528,26 @@ class ReahlWebServer(object):
         self.reahl_wsgi_app.stop()
         self.httpd.server_close()
         self.httpsd.server_close()
-    
-    def requests_waiting(self, timeout):
-        return self.httpd.requests_waiting(timeout) or self.httpsd.requests_waiting(timeout/10)
+
+    def connection_is_pending(self, timeout):
+        return self.httpd.connection_is_pending(timeout) or self.httpsd.connection_is_pending(timeout/10)
 
     def serve_until(self, done):
         while not (done() or self.reahl_wsgi_app.has_uncaught_exception()):
-            self.httpd.serve_async()
-            self.httpsd.serve_async()
+            self.httpd.serve_async(in_separate_thread=self.in_separate_thread)
+            self.httpsd.serve_async(in_separate_thread=self.in_separate_thread)
 
     def serve(self, timeout=0.01):
         """Call this method once to have the server handle all waiting requests in the calling thread."""
         def done():
-            return not self.requests_waiting(timeout)
+            return not self.connection_is_pending(timeout)
         self.serve_until(done)
 
     def install_handler(self, web_driver):
         """Installs this server's request handler into the given `web_driver`. This enables the
            server to serve requests from the web_driver in the current thread."""
-        assert web_driver not in self.handlers.keys(), 'Handler already installed into %s' % web_driver
-        new_handler = Handler(web_driver.command_executor)
+        assert web_driver not in self.handlers.keys(), 'WebDriverHandler already installed into %s' % web_driver
+        new_handler = WebDriverHandler(web_driver.command_executor)
         self.handlers[web_driver] = new_handler
         new_handler.install(self)
 
@@ -363,6 +559,18 @@ class ReahlWebServer(object):
         for handler in self.handlers.values():
             handler.reinstall()
 
+    @contextmanager
+    def paused(self, wait_till_done_serving=True):
+        self.restore_handlers()
+        try:
+            yield
+        finally:
+            try:
+                if wait_till_done_serving:
+                    self.serve()
+            finally:
+                self.reinstall_handlers()
+
     @contextmanager
     def in_background(self, wait_till_done_serving=True):
         """Returns a context manager. Within the context of this context manager, the webserver is temporarily run
